# -*- coding: utf-8 -*-

from numpy import median

class SubCluster():
    ## cluster is the component in terms of
    ## nodes and neighbor_link
    def __init__(self, cid, nodeslist):
        self.cluster_id = cid
        self.node_list = nodeslist
        self.cluster_size = len(nodeslist)
        self.xcor = None
        self.ycor = None
        self.time_start = None
        self.time_stop = None
        self.time_median = None
        self.chain_id = None
        self._calculation_info()
        self.outgoings = []
        self.incomings = []
        self.behaviors = []

    def _calculation_info(self):
        xlist = [n.xcor for n in self.node_list]
        ylist = [n.ycor for n in self.node_list]
        self.xcor = sum(xlist)/float(len(xlist))
        self.ycor = sum(ylist)/float(len(ylist))

        tlist = [n.time for n in self.node_list]
        self.time_start = min(tlist)
        self.time_stop = max(tlist)
        self.time_median = median(tlist)

    def add_outgoing(self, acluster):
        self.outgoings.append(acluster)

    def add_incoming(self, acluster):
        self.incomings.append(acluster)

    def set_chain_id(self, cid):
        self.chain_id = cid

    def finalize_behaviors(self):
        if len(self.incomings)<1:
            self.behaviors.append('appearing')
        elif len(self.incomings)>1:
            self.behaviors.append('merging')
        else: ## ==1
            c0 = self.incomings[0]
            l0 = c0.cluster_size
            l1 = self.cluster_size
            if l0>l1:
                self.behaviors.append('shrinking')
            elif l0<l1:
                self.behaviors.append('growing')
            else:
                self.behaviors.append('maintained')
        if len(self.outgoings)<1:
            self.behaviors.append('disappearing')
        elif len(self.outgoings)>1:
            self.behaviors.append('splitting')


"""
class Component_Graph():
    ## component_graph is the component in terms of
    ## clusters and linked by their nodes' shifting_link
    def __init__(self):
        pass
"""



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Polygon, LineString, Point
from shapely.geometry.polygon import LinearRing # get ellipse

def get_dist(pt0, pt1):
    x0, y0 = pt0
    x1, y1 = pt1
    d = np.sqrt((x1-x0)**2+(y1-y0)**2)
    return d

class Output_Objects(object):
    def __init__(self, parent):
        self.parent = parent
        self.node_df = self.get_node_df()
        self.cluster_df = self.get_cluster_df()
        self.progress_df = self.get_progress_df()
        self.summary_df = self.get_summary_df()
        self.final_slinks_df = self.get_final_slinks_df()
        self.final_nlinks_df = self.get_final_nlinks_df()

    def get_summary_df(self):
        ## working 11/04
        nodes = self.parent.sg.nodes
        nlinks = self.parent.sg.nlinks
        slinks = self.parent.sg.slinks
        final_slinks = self.parent.sg.slinks_final
        cluster_pairs = self.parent.sg.nlinks_final
        clusters = self.parent.pg.clusters
        lines = self.parent.pg.progress
        cv = self.parent.critical_value
        dic = dict(
            nodes=len(nodes),
            npair=len(nlinks),
            slink=len(slinks),
            final_cpair=len(cluster_pairs),
            final_slink=len(final_slinks),
            clusterno=len(clusters),
            progressno=len(lines),
            critical_value=cv)
        df_summary = pd.DataFrame.from_dict(dict(attribute=dic))
        return df_summary

    def get_node_df(self):
        nodes = self.parent.sg.nodes
        ii = []#[c.idd for c in nodes]
        xx = []#[c.xcor for c in nodes]
        yy = []#[c.ycor for c in nodes]
        tt = []#[c.time for c in nodes]
        cl = []#[c.cluster_id for c in nodes]
        ch = []#[c.chain_id for c in nodes]
        nc = []#[len(c.incoming) for c in nodes]
        ou = []#[len(c.outgoing) for c in nodes]
        ne = []#[len(c.neighbors) for c in nodes]
        for n in nodes:
            ii.append(n.idd)
            xx.append(n.xcor)
            yy.append(n.ycor)
            tt.append(n.time)
            cl.append(n.cluster_id)
            ch.append(n.chain_id)
            nc.append(len(n.incoming))
            ou.append(len(n.outgoing))
            ne.append(len(n.neighbors))

        df_nodes = pd.DataFrame.from_dict({'node_id':ii, 'xx':xx, 'yy':yy, 'time':tt, 'subcluster_id':cl, 'chain_id':ch, 'in_size':nc, 'out_size':ou, 'neig_size':ne})
        df_nodes = df_nodes[['node_id', 'xx', 'yy', 'time', 'subcluster_id', 'chain_id', 'in_size', 'out_size', 'neig_size']]
        #self.node_df = df_nodes
        return df_nodes

    def get_cluster_df(self):
        clusters = self.parent.pg.clusters
        xx = []#[c.xcor for c in clusters]
        yy = []#[c.ycor for c in clusters]
        ss = []#[c.cluster_size for c in clusters]
        tt = []#[c.time_median for c in clusters]
        t0 = []#[c.time_start for c in clusters]
        t1 = []#[c.time_stop for c in clusters]
        cl = []#[c.cluster_id for c in clusters]
        ch = []#[c.chain_id for c in clusters]
        bvr = []
        ins = []
        ous = []
        for c in clusters:
            xx.append(c.xcor)
            yy.append(c.ycor)
            ss.append(c.cluster_size)
            tt.append(c.time_median)
            t0.append(c.time_start)
            t1.append(c.time_stop)
            cl.append(c.cluster_id)
            ch.append(c.chain_id)
            bvr.append(c.behaviors)
            ins.append(len(c.incomings))
            ous.append(len(c.outgoings))
        df_clusters = pd.DataFrame.from_dict({'subcluster_id':cl, 'chain_id':ch, 'xx':xx, 'yy':yy, 'cluster_size':ss, 'time_median':tt, 'time_start':t0, 'time_stop':t1, 'behaviors':bvr, 'in_count':ins, 'out_count':ous})
        df_clusters = df_clusters[['subcluster_id', 'chain_id', 'xx', 'yy', 'cluster_size', 'time_median', 'time_start', 'time_stop', 'behaviors', 'in_count', 'out_count']]
        #self.cluster_df = df_clusters
        return df_clusters

    def get_progress_df(self):
        #progress = self.pg.progress
        lines = self.parent.pg.progress
        ii = 0
        lines_dict = {}
        for line in lines:
            i0 = line.origin.cluster_id
            i1 = line.destination.cluster_id
            s0 = line.origin.cluster_size
            s1 = line.destination.cluster_size
            x0 = line.origin.xcor
            x1 = line.destination.xcor
            y0 = line.origin.ycor
            y1 = line.destination.ycor
            z0 = line.origin.time_start
            z1 = line.destination.time_stop
            clid = line.destination.cluster_id
            chid = line.destination.chain_id
            op = line.ori_pos
            no_SL = line.no_slink
            dist = get_dist((x0,y0),(x1,y1))
            lag = z1 - z0
            lines_dict[ii] = {'id0':i0, 'id1':i1, 'clid':clid, 'chid':chid, 'size0':s0, 'size1':s1, 'x0':x0, 'x1':x1, 'y0':y0, 'y1':y1, 't0':z0, 't1':z1, 'op':op, 'no_SL':no_SL, 'distance':dist, 'timelag':lag}
            ii=ii+1
        if len(lines)<=0:
            lines_dict[ii] = {'id0':'-', 'id1':'-', 'clid':'-', 'chid':'-', 'size0':'-', 'size1':'-', 'x0':'-', 'x1':'-', 'y0':'-', 'y1':'-',
            't0':'-',  't1':'-',  'op':'-', 'no_SL':'-', 'distance':'-', 'timelag':'-'}
        df_progress = pd.DataFrame.from_dict(lines_dict, orient='index')
        df_progress = df_progress[['id0','id1','clid','chid','size0','size1','x0','x1','y0','y1','t0','t1','op','no_SL', 'distance','timelag']]
        #self.progress_df = df_progress
        return df_progress

    def get_final_slinks_df(self):
        final_slinks = self.parent.sg.slinks_final
        final_dict = {}
        i = 0
        if len(final_slinks)>0:
            for sl in final_slinks:
                oo = sl.origin
                dd = sl.destination
                sldic = dict(
                    ooid = oo.idd,
                    ddid = dd.idd,
                    srisk = sl.spatial_risk,
                    trisk = sl.temporal_risk,
                    crisk = sl.combine_risk,
                    opossi = sl.ori_possibility,
                    oxcor = oo.xcor,
                    oycor = oo.ycor,
                    otime = oo.time,
                    dxcor = dd.xcor,
                    dycor = dd.ycor,
                    dtime = dd.time,
                    clid = dd.cluster_id,
                    chid = dd.chain_id,
                    distance = get_dist((oo.xcor,oo.ycor),(dd.xcor,dd.ycor)),
                    timelag = dd.time - oo.time,
                )
                final_dict[i] = sldic
                i+=1
        else:
            sldic = dict(
                ooid = '-',
                ddid = '-',
                srisk = '-',
                trisk = '-',
                crisk = '-',
                opossi = '-',
                oxcor = '-',
                oycor = '-',
                otime = '-',
                dxcor = '-',
                dycor = '-',
                dtime = '-',
                clid = '-',
                chid = '-',
                distance = '-',
                timelag = '-',
            )
            final_dict[i] = sldic
        f_sls = pd.DataFrame.from_dict(final_dict, orient='index')
        f_sls = f_sls[['ooid','ddid','clid','chid','srisk','trisk','crisk','opossi','oxcor','oycor','otime','dxcor','dycor','dtime', 'distance', 'timelag']]
        return f_sls

    def get_final_nlinks_df(self):
        cluster_pairs = self.parent.sg.nlinks_final
        final_dict = {}
        i = 0
        if len(cluster_pairs)>0:
            for cp in cluster_pairs:
                n1 = cp.one
                n2 = cp.two
                cpdic = dict(
                    n1_id = n1.idd,
                    n2_id = n2.idd,
                    clid = n1.cluster_id,
                    chid = n1.chain_id,
                    max_cop = cp.max_cop,
                    n1x = n1.xcor,
                    n1y = n1.ycor,
                    n2x = n2.xcor,
                    n2y = n2.ycor,
                    n1t = n1.time,
                    n2t = n2.time,
                    distance = get_dist((n1.xcor,n1.ycor),(n2.xcor,n2.ycor)),
                    timelag = n2.time - n1.time,
                )
                final_dict[i] = cpdic
                i+=1
        else:
            cpdic = dict(
                n1_id = '-',
                n2_id = '-',
                clid = '-',
                chid = '-',
                max_cop = '-',
                n1x = '-',
                n1y = '-',
                n2x = '-',
                n2y = '-',
                n1t = '-',
                n2t = '-',
                distance = '-',
                timelag = '-',
            )
            final_dict[i] = cpdic
        f_nps = pd.DataFrame.from_dict(final_dict, orient='index')
        f_nps = f_nps[['n1_id', 'n2_id', 'clid', 'chid', 'max_cop','n1x','n1y','n2x','n2y', 'n1t', 'n2t', 'distance', 'timelag']]
        return f_nps

    def get_node_gdf(self):
        ndf = self.get_node_df()
        xx = ndf.xx.tolist()
        yy = ndf.yy.tolist()
        geom = []
        for a,b in zip(xx,yy):
            p = Point(a,b)
            geom.append(p)
        geo_df = gpd.GeoDataFrame(ndf, crs=self.parent.crs, geometry=geom)
        return geo_df

    def get_cluster_gdf(self, vno=16, dev_scale=1.5):
        ndf = self.get_node_df()
        cdf = self.get_cluster_df()
        cdf.sort_values(by='chain_id', inplace=True)
        cids2 = cdf.chain_id.tolist()
        clids = cdf.cluster_id.tolist()
        cluster_ellipse = []
        cluster_area = []
        for cl in clids:
            temp_df = ndf[ndf.cluster_id==cl]
            temp_x = temp_df.xx.tolist()
            temp_y = temp_df.yy.tolist()
            ellipse = self.__get_ellipse(temp_x, temp_y, vno, dev_scale)
            poly = Polygon(ellipse)
            cluster_ellipse.append(poly)
            cluster_area.append(poly.area)
        beh = cdf.behaviors.tolist()
        cdf['behaviors'] = [ str(b) for b in beh ]
        new_df = cdf.rename(columns={'cluster_size':'cls_size', 'time_median':'time_mdian'})
        geo_df = gpd.GeoDataFrame(new_df, crs=self.parent.crs, geometry=cluster_ellipse)
        geo_df['cls_area'] = cluster_area
        return geo_df

    def __get_ellipse(self, xlist, ylist, vno, dev_scale):
        # vno is the number of vertices of the ellipse
        # dev_scale is the standard deviation scale
        size = len(xlist)
        xx = sum(xlist)/float(size)
        yy = sum(ylist)/float(size)
        xwave = [(a-xx) for a in xlist]
        ywave = [(b-yy) for b in ylist]
        xlist2 = [a**2 for a in xwave]
        ylist2 = [b**2 for b in ywave]
        xylist2 = [(a)*(b) for a,b in zip(xwave,ywave)]
        sumx2 = sum(xlist2)
        sumy2 = sum(ylist2)
        sumxy2 = sum(xylist2)
        A = sumx2 - sumy2
        B = np.sqrt((sumx2 - sumy2)**2 + 4*(sumxy2)**2)
        C = 2*sumxy2
        #circle = []
        if C!=0:
            theta = np.arctan((A+B)/C)
            cxlist = [a*np.cos(theta) for a in xwave]
            sylist = [b*np.sin(theta) for b in ywave]
            sxlist = [a*np.sin(theta) for a in xwave]
            cylist = [b*np.cos(theta) for b in ywave]
            partx = [(a-b)**2 for a,b in zip(cxlist,sylist)]
            party = [(a-b)**2 for a,b in zip(sxlist,cylist)]
            devx = np.sqrt(sum(partx)/float(size))*dev_scale
            devy = np.sqrt(sum(party)/float(size))*dev_scale
            #print devx

            n = 64
            t = np.linspace(0, 2*np.pi, n, endpoint=False)
            st = np.sin(t)
            ct = np.cos(t)
            sa = np.sin(theta)
            ca = np.cos(theta)
            p = np.empty((n, 2))
            p[:, 0] = xx + devx * ca * ct - devy * sa * st
            p[:, 1] = yy + devx * sa * ct + devy * ca * st
            circle = list(LinearRing(p).coords)
            #circle = LinearRing(p)
        else:
            dev = (np.sqrt(sum(xlist2)/float(size)) + 1)*dev_scale
            #print dev
            theta = 1
            n = 64
            t = np.linspace(0, 2*np.pi, n, endpoint=False)
            st = np.sin(t)
            ct = np.cos(t)
            sa = np.sin(theta)
            ca = np.cos(theta)
            p = np.empty((n, 2))
            p[:, 0] = xx + dev * ca * ct - dev * sa * st
            p[:, 1] = yy + dev * sa * ct + dev * ca * st
            circle = list(LinearRing(p).coords)
            #circle = LinearRing(p)
        #print circle
        skip = int(np.floor(len(circle[:-1])/float(vno-1)))
        temp = []
        #print vno
        for i in range(vno):
            j = skip*i
            #print j
            temp.append(circle[j])
        return circle

    def get_progress_gdf(self):
        #outputfilename = filename_prefix+'_proglinks.shp'
        edf = self.get_progress_df()
        if not (edf.iloc[0]['x0']=='-'):
            geom = []
            #dist = []
            #lags = []
            for i in range(len(edf)):
                row = edf.iloc[i]
                x0 = row['x0']
                x1 = row['x1']
                y0 = row['y0']
                y1 = row['y1']
                #t0 = row['t0']
                #t1 = row['t1']
                #lags.append(float(t1)-float(t0))
                #dist.append(get_dist((x0,y0),(x1,y1)))
                line = LineString(((x0,y0),(x1,y1)))
                geom.append(line)
            #crs = '+proj=tmerc +lat_0=0 +lon_0=121 +k=0.9999 +x_0=250000 +y_0=0 +ellps=GRS80 +units=m +no_defs'
            #edf['time_lag'] = lags
            #edf['distance'] = dist
            geo_df = gpd.GeoDataFrame(edf, crs=self.parent.crs, geometry=geom)
            #print geo_df.head()
            return geo_df
        else:
            pass

    def get_final_slinks_gdf(self):
        #outputfilename = filename_prefix+'_slinks_final.shp'
        edf = self.get_final_slinks_df()
        if not (edf.iloc[0]['ooid']=='-'):
            geom = []
            #dist = []
            #lags = []
            for i in range(len(edf)):
                row = edf.iloc[i]
                x0 = row['oxcor']
                x1 = row['dxcor']
                y0 = row['oycor']
                y1 = row['dycor']
                #t0 = row['otime']
                #t1 = row['dtime']
                #lags.append(float(t1)-float(t0))
                #dist.append(get_dist((x0,y0),(x1,y1)))
                line = LineString(((x0,y0),(x1,y1)))
                geom.append(line)
            #edf['time_lag'] = lags
            #edf['distance'] = dist
            geo_df = gpd.GeoDataFrame(edf, crs=self.parent.crs, geometry=geom)
            #print geo_df.head()
            return geo_df
        else:
            pass

    def get_final_nlinks_gdf(self):
        #outputfilename = filename_prefix+'_cluster_pairs.shp'
        edf = self.get_final_nlinks_df()
        if not (edf.iloc[0]['n1_id']=='-'):
            geom = []
            #dist = []
            #lags = []
            for i in range(len(edf)):
                row = edf.iloc[i]
                x0 = row['n1x']
                x1 = row['n2x']
                y0 = row['n1y']
                y1 = row['n2y']
                #t0 = row['n1t']
                #t1 = row['n2t']
                #lags.append(float(t1)-float(t0))
                #dist.append(get_dist((x0,y0),(x1,y1)))
                line = LineString(((x0,y0),(x1,y1)))
                geom.append(line)
            #edf['time_lag'] = lags
            #edf['distance'] = dist
            geo_df = gpd.GeoDataFrame(edf, crs=self.parent.crs, geometry=geom)
            #print geo_df.head()
            return geo_df
        else:
            pass



# -*- coding: utf-8 -*-
import os
from datetime import datetime, timedelta
import pandas as pd
import geopandas as gpd
from descartes import PolygonPatch # for drawing base polygon

def to_anime(oo, filename_prefix=None, bg_polys=[], bbox=None):
    # oo == Output_Objects
    if filename_prefix is None:
        filename_prefix = 'temp_output/temp'

    dirname = os.path.dirname(filename_prefix)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    to_3plots(oo, filename_prefix, bg_polys, bbox)

def to_3plots(oo, filename_prefix, bg_polys, bbox):
    # animation with 3 subplots
    export_file_name = filename_prefix+"_compare_movie.mp4"
    get_anim(oo, export_file_name, bg_polys, bbox)

def get_cases(oo):
    lag_disappear = 12
    df2 = oo.node_df
    #print df2.head()
    ii = df2.node_id.tolist()
    xx = df2.xx.tolist()
    yy = df2.yy.tolist()
    tt1 = df2.time.tolist()
    tt2 = [ t+lag_disappear for t in tt1 ]
    cases_a = {}
    cases_b = {}
    cases = {}
    for a,b,c,d,e in zip(ii,xx,yy,tt1,tt2):
        cases[a] = (b,c)
        if d not in cases_a:
            cases_a[d] = []
        cases_a[d].append(a)
        if e not in cases_b:
            cases_b[e] = []
        cases_b[e].append(a)
    minx = min(xx)
    maxx = max(xx)
    miny = min(yy)
    maxy = max(yy)
    bbox = [minx,miny,maxx,maxy]
    return cases, cases_a, cases_b, bbox

def get_subclusters(oo, filename_prefix):
    #sfile = 'exports/new5/c05_y'+str(y)+'_result_clusters.csv'
    filename_prefix2 = filename_prefix.replace("_compare_movie.mp4","")
    efile = filename_prefix2+'_ellipse.shp'
    edf = None
    if os.path.exists(efile):
        edf = gpd.read_file(efile)
    else:
        import to_shpfile
        dirname = os.path.dirname(filename_prefix)
        temp_prefx = dirname+'/temp/temp'
        dirname2 = os.path.dirname(temp_prefx)
        if not os.path.isdir(dirname2):
            os.makedirs(dirname2)
        to_shpfile.to_eclipse(oo, filename_prefix=temp_prefx, crs=None, vno=16, dev_scale=1.5)
        edf = gpd.read_file(temp_prefx+'_ellipse.shp')

    ii = edf.cluster_id.tolist()
    geoms = edf.geometry.tolist()
    geom_dict = {}
    cent_dict = {}
    for k,v in zip(ii,geoms):
        geom_dict[k] = v
        a,b,c,d = v.bounds
        x = float(c - a)/2. + a
        y = float(d - b)/2. + b
        cent_dict[k] = (x,y)
        #print x,y

    t1 = edf.time_start.tolist()
    t2 = edf.time_stop.tolist()
    time_dict = { k:(a,b) for k,a,b in zip(ii,t1,t2) }
    #print edf2
    sub_a = {}
    sub_b = {}
    for k,a,b in zip(ii,t1,t2):
        if a not in sub_a:
            sub_a[a] = []
        if b not in sub_b:
            sub_b[b] = []
        sub_a[a].append(k)
        sub_b[b].append(k)
    return geom_dict, time_dict, sub_a, sub_b, cent_dict

def get_chains(oo):
    global time_dict, cent_dict
    df2 = oo.progress_df
    #print df2.head()
    i0 = df2.id0.tolist()
    i1 = df2.id1.tolist()
    x0 = []#df2.x0.tolist()
    x1 = []#df2.x1.tolist()
    y0 = []#df2.y0.tolist()
    y1 = []#df2.y1.tolist()
    t0 = []
    t1 = []
    for i in i1:
        a,b = time_dict[i]
        t0.append(a)
        t1.append(b)
        xx,yy = cent_dict[i]
        x1.append(xx)
        y1.append(yy)
    for i in i0:
        xx,yy = cent_dict[i]
        x0.append(xx)
        y0.append(yy)
    #t0 = df2.t0.tolist()
    #t1 = df2.t1.tolist()
    arrows = []
    start_arrows = {}
    stop_arrows = {}
    sets = zip(x0,x1,y0,y1,t0,t1)
    for i in range(len(sets)):
        a,b,c,d,e,f = sets[i]
        if a!='-':
            arrows.append(([a,b],[c,d]))
            if e>f:
                print(e,f)
            e = int(e)
            f = int(f)
            if e not in start_arrows:
                start_arrows[e] = []
            if f not in stop_arrows:
                stop_arrows[f] = []
            start_arrows[e].append(i)
            stop_arrows[f].append(i)
    return (arrows, start_arrows, stop_arrows)


def update(i):
    global cases, cases_a, cases_b, geom_dict, time_dict, sub_a, sub_b, cent_dict, arrows, arrows0, arrows1, time_text, ax, ax2, ax3, drawed, zup, zdown
    #global y, arrows, arrows0, arrows1, time_text, ax, ax2, ax3, current, drawed
    #this_color = colors[16]
    time_text.set_text('time: '+str(i))
    day = i

    if day in cases_a:
        for j in cases_a[day]:
            xc,yc = cases[j]
            drawed['case'][j] = ax.scatter(xc,yc, s=10, c='#f7022a', edgecolors='#f7022a', zorder=zup, alpha=0.8)

    if day in cases_b:
        for j in cases_b[day]:
            p = drawed['case'][j]
            p.remove()

            xc,yc = cases[j]
            p2 = ax.scatter(xc,yc, s=10, c='#fdc1c5', edgecolors='#fdc1c5', zorder=zdown, alpha=0.6)

    if day in sub_a:
        for j in sub_a[day]:
            gg = geom_dict[j]
            g = ax2.add_patch(PolygonPatch(gg, fc='#f7022a', lw=1.2, ec='#f7022a', alpha=0.6, zorder=zup ))
            drawed['subcluster'][j] = g

    if day in sub_b:
        for j in sub_b[day]:
            g = drawed['subcluster'][j]
            g.remove()

            gg = geom_dict[j]
            g2 = ax2.add_patch(PolygonPatch(gg, fc='#fdc1c5', lw=1.2, ec='#fdc1c5', alpha=0.5, zorder=zdown ))


    if day in arrows0:
        for j in arrows0[day]:
            #print j
            (x0,x1),(y0,y1) = arrows[j]
            drawed['arrow'][j] = ax3.arrow(x0,y0,x1-x0,y1-y0, linewidth=1.5, head_width=100., head_length=150., fc='#f7022a', ec='#f7022a', zorder=zup, alpha=0.9)

    if day in arrows1:
        for j in arrows1[day]:
            a = drawed['arrow'][j]
            a.remove()

            (x0,x1),(y0,y1) = arrows[j]
            b = ax3.arrow(x0,y0,x1-x0,y1-y0, linewidth=1.5, head_width=100., head_length=150., fc='#fdc1c5', ec='#fdc1c5', zorder=zdown, alpha=0.6)
            #prelines.append(b)

def get_anim(oo, filename_prefix, bg_polys, bbox):
    import matplotlib.pyplot as plt
    import seaborn as sns
    from matplotlib.animation import FuncAnimation
    global cases, cases_a, cases_b, geom_dict, time_dict, sub_a, sub_b, cent_dict, arrows, arrows0, arrows1, time_text, ax, ax2, ax3, drawed, zup, zdown
    #minx,maxx = 175000,186000
    #miny,maxy = 2495000,2506000
    print("start exporting movie")

    fig,(ax,ax2,ax3) = plt.subplots(1,3, figsize=(16,6), sharex=True, sharey=True)
    ax.set_aspect("equal")
    ax2.set_aspect("equal")
    ax3.set_aspect("equal")
    time_text = fig.suptitle('', fontsize=18, fontweight='bold')
    ax.set_xlabel("Case", fontsize=18)
    ax2.set_xlabel("Sub-clusters", fontsize=18)
    ax3.set_xlabel("Progression links", fontsize=18)
    plt.tight_layout()
    plt.subplots_adjust(top=0.9)
    #time_text = ax.text(0.02, 0.02, '', transform=ax.transAxes)
    #ax2.set_aspect("equal")
    bg = 'black'

    zo = 2
    for bgset in bg_polys:
        if len(bgset)==3:
            bgf, bgc, bgw = bgset
            bga = 0.6
        elif len(bgset)==4:
            bgf, bgc, bgw, bga = bgset
        else:
            print("background items should be a tuple with len==3 or 4")
        gdf = gpd.read_file(bgf)
        for i in range(len(gdf)):
            poly = gdf['geometry'][i]
            ax.add_patch(PolygonPatch(poly, fill=False, lw=bgw, ec=bgc, alpha=bga, zorder=zo ))
            ax2.add_patch(PolygonPatch(poly, fill=False, lw=bgw, ec=bgc, alpha=bga, zorder=zo ))
            ax3.add_patch(PolygonPatch(poly, fill=False, lw=bgw, ec=bgc, alpha=bga, zorder=zo ))
        zo+=1
    zdown = zo
    zup = zo + 1
    cases, cases_a, cases_b, bbox2 = get_cases(oo)
    geom_dict, time_dict, sub_a, sub_b, cent_dict = get_subclusters(oo, filename_prefix)
    arrows,arrows0,arrows1 = get_chains(oo)
    if bbox is None:
        bbox = bbox2
    minx,miny,maxx,maxy = bbox
    """
    for xs,ys in arrows:
        x0,x1 = xs
        y0,y1 = ys
        ax.arrow(x0,y0,x1-x0,y1-y0, linewidth=3, head_width=100., head_length=150., fc='#f7022a', ec='#f7022a', zorder=3, alpha=0.1)
    """
    drawed = {}
    drawed['case'] = {}
    drawed['subcluster'] = {}
    drawed['arrow'] = {}

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_xlim((minx,maxx))
    ax.set_ylim((miny,maxy))

    anim = FuncAnimation(fig, update, frames=365, interval=2000)
    anim.save(filename_prefix, fps=8, extra_args=['-vcodec', 'libx264'])
    #plt.show()
    plt.close()
    print("movie exported")



# -*- coding: utf-8 -*-

from .subcluster import SubCluster
from .progress_link import Progress_Link


class Progression_Graph():
    ## the total picture, and the combinations of all clusters
    def __init__(self):
        self.clusters = []
        self.clusters_dict = {}
        self.progress = []
        self.progress_pair = []
        self.prog_dict = {}

    #def get_clusters(self):
    #    newlist = sorted(self.clusters, key=lambda x: x.time_start, reverse=False)
    #    return newlist

    def add_cluster_node(self, j, nodes):
        a_cluster = SubCluster(j, nodes)
        self.clusters.append(a_cluster)
        self.clusters_dict[j] = a_cluster

    def finalize_clusters(self):
        self.clusters.sort(key=lambda x: x.time_start, reverse=False)

    def add_progressionlink(self, cid1, cid2, op):
        #cid1 = cluster1.cluster_id
        #cid2 = cluster2.cluster_id
        cluster1 = self.clusters_dict[cid1]
        cluster2 = self.clusters_dict[cid2]
        #if (cid1,cid2) not in self.progress_pair:
        if cluster2 not in cluster1.outgoings:
            PL = Progress_Link(cluster1, cluster2, op)
            self.progress.append(PL)
            self.progress_pair.append((cid1,cid2))
            self.prog_dict[(cid1,cid2)] = PL
            cluster1.add_outgoing(cluster2)
            cluster2.add_incoming(cluster1)
        else:
            PL = self.prog_dict[(cid1,cid2)]
            PL.set_op(op)



# -*- coding: utf-8 -*-

class Progress_Link():
    ## the directed links between clusters in each progression_graph
    def __init__(self, cluster1, cluster2, op):
        self.origin = cluster1
        self.destination = cluster2
        self.ori_pos = op
        self.no_slink = 1

    def set_op(self, ori_pos):
        if ori_pos>self.ori_pos:
            self.ori_pos = ori_pos
        self._add_no_slink()

    def _add_no_slink(self):
        self.no_slink += 1



# -*- coding: utf-8 -*-

from .output_objects import Output_Objects
from .to_csv import to_csv
from .to_shpfile import to_shpfile
from .to_anime import to_anime

"""
### svg output functions are still in experiment
from .to_svgs import to_svgs
from .to_svgs import to_svgs_all
from .to_svgs import to_svg_event
from .to_svgs import to_svg_animate
from .to_svgs import to_svg_static
from .to_svgs import to_svg_morph
"""



# -*- coding: utf-8 -*-

from .node import Node
from .neighbor_pair import Neighbor_Pair
from .shifting_link import Shifting_Link
from .shift_graph import Shift_Graph

from .subcluster import SubCluster
from .progress_link import Progress_Link
from .progression_graph import Progression_Graph



# -*- coding: utf-8 -*-

### svg output functions are still in experiment

import os
from shutil import copyfile
import numpy as np
from shapely.geometry.polygon import LinearRing
from shapely.geometry import Point

"""
from vmapper import make_svg as svg_making
from vmapper import svg
from vmapper import fromShpfile
"""

def to_svgs(oo, filename_prefix=None, pgraph=None, bgfile=None, bbox=None, scale=30., widthscale=1.):
    pass
"""
def to_svgs(oo, filename_prefix=None, pgraph=None, bgfile=None, bbox=None, scale=30., widthscale=1.):
    to_svgs_all(oo, filename_prefix, pgraph, bgfile, bbox, scale, widthscale)

def to_svgs_all(oo, filename_prefix=None, pgraph=None, bgfile=None, bbox=None, scale=30., widthscale=1.):
    # oo == Output_Objects

    to_svg_event(oo, bgfile, filename_prefix, title="The event of points", bbox=bbox, scale=scale)
    to_svg_animate(oo, pgraph, bgfile, filename_prefix, title="The progression of subcluster", bbox=bbox, scale=scale)
    to_svg_static(oo, pgraph, bgfile, filename_prefix, title="The static view of subcluster", bbox=bbox, shape='ellipse', widthscale=widthscale)
    to_svg_morph(oo, pgraph, bgfile, filename_prefix, title="The progression of subcluster", bbox=bbox, shape='ellipse', scale=scale)

def to_svg_event(oo, bgfile, filename_prefix, title="The event of points", bbox=None, scale=30.):
    if bbox is None:
        bbox = [None, None, None, None]
    xmin,ymin,xmax,ymax = bbox
    event_to_animate_svg(oo, bgfile, filename_prefix, title=title, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, scale=scale)

def to_svg_animate(oo, pgraph, bgfile, filename_prefix, title="The progression of subcluster", bbox=None, scale=30.):
    if bbox is None:
        bbox = [None, None, None, None]
    xmin,ymin,xmax,ymax = bbox
    pg_to_animate_svg(oo, pgraph, bgfile, filename_prefix, title=title, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, scale=scale)

def to_svg_static(oo, pgraph, bgfile, filename_prefix, title="The static view of subcluster", bbox=None, shape='ellipse', widthscale=1.):
    if bbox is None:
        bbox = [None, None, None, None]
    xmin,ymin,xmax,ymax = bbox
    pg_to_static_svg(oo, pgraph, bgfile, filename_prefix, title=title, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, shape=shape, widthscale=widthscale)

def to_svg_morph(oo, pgraph, bgfile, filename_prefix, title="The progression of subcluster", bbox=None, shape='ellipse', scale=30.):
    if bbox is None:
        bbox = [None, None, None, None]
    xmin,ymin,xmax,ymax = bbox
    pg_to_morph_svg(oo, pgraph, bgfile, filename_prefix, title=title, xmin=xmin, ymin=ymin, xmax=xmax, ymax=ymax, shape=shape, scale=scale)

def _get_background(bgf, layername=None, labelby=None,
    dcolor=(5,5,5), dopacity=0.7, dscolor=(255,255,255), dswidth=5, widthscale=1.):
    if not(bgf is None):
        ss = fromShpfile.fromShp(bgf)
        shps = ss.shapedict
        bbox = ss.bbox()
        layer = svg.Layer()
        layer.addtoLayer(svg.initLayer(key=layername, dcolor=dcolor, dopacity=dopacity,
            dscolor=dscolor, dswidth=dswidth*widthscale))
        for shp_key, shp_val in shps.iteritems():
            if labelby is not None:
                label = labels[shp_key]
            else:
                label = shp_key
            layer.addtoLayer(svg.MultiPolygons(key=shp_key, label=label, multiPolygons=shp_val))
        layer.addtoLayer(svg.closeLayer())
        return (layer.items, bbox)
    else:
        return None

def _get_circle(node_list, vno=16, scale=30.):
    size = len(node_list)
    xlist = [a.xcor for a in node_list]
    ylist = [a.ycor for a in node_list]
    xx = sum(xlist)/float(size)
    yy = sum(ylist)/float(size)
    circle = list(Point(xx,yy).buffer(size*scale).exterior.coords)
    temp = []
    skip = int(np.floor(len(circle[:-1])/float(vno-1)))
    for i in range(vno):
        j = skip*i
        temp.append(circle[j])
    temp.append(temp[0])
    #print len(temp)
    return temp

def _get_ellipse(node_list, vno=16, dev_scale=1.5):
    size = len(node_list)
    xlist = [a.xcor for a in node_list]
    ylist = [a.ycor for a in node_list]
    xx = sum(xlist)/float(size)
    yy = sum(ylist)/float(size)

    xwave = [(a-xx) for a in xlist]
    ywave = [(b-yy) for b in ylist]
    xlist2 = [a**2 for a in xwave]
    ylist2 = [b**2 for a in ywave]
    xylist2 = [(a)*(b) for a,b in zip(xwave,ywave)]
    sumx2 = sum(xlist2)
    sumy2 = sum(ylist2)
    sumxy2 = sum(xylist2)
    A = sumx2 - sumy2
    B = np.sqrt((sumx2 - sumy2)**2 + 4*(sumxy2)**2)
    C = 2*sumxy2
    circle = []
    if C != 0:
        theta = np.arctan((A+B)/C)

        cxlist = [a*np.cos(theta) for a in xwave]
        sylist = [b*np.sin(theta) for b in ywave]
        sxlist = [a*np.sin(theta) for a in xwave]
        cylist = [b*np.cos(theta) for b in ywave]
        partx = [(a-b)**2 for a,b in zip(cxlist,sylist)]
        party = [(a-b)**2 for a,b in zip(sxlist,cylist)]
        devx = np.sqrt(sum(partx)/float(size))*dev_scale
        devy = np.sqrt(sum(party)/float(size))*dev_scale
        #print devx

        n = 64
        t = np.linspace(0, 2*np.pi, n, endpoint=False)
        st = np.sin(t)
        ct = np.cos(t)
        sa = np.sin(theta)
        ca = np.cos(theta)
        p = np.empty((n, 2))
        p[:, 0] = xx + devx * ca * ct - devy * sa * st
        p[:, 1] = yy + devx * sa * ct + devy * ca * st
        circle = list(LinearRing(p).coords)
    else:
        dev = (np.sqrt(sum(xlist2)/float(size)) + 1)*dev_scale
        #print dev
        theta = 1
        n = 64
        t = np.linspace(0, 2*np.pi, n, endpoint=False)
        st = np.sin(t)
        ct = np.cos(t)
        sa = np.sin(theta)
        ca = np.cos(theta)
        p = np.empty((n, 2))
        p[:, 0] = xx + dev * ca * ct - dev * sa * st
        p[:, 1] = yy + dev * sa * ct + dev * ca * st
        circle = list(LinearRing(p).coords)

    #print circle
    skip = int(np.floor(len(circle[:-1])/float(vno-1)))
    temp = []
    #print vno
    for i in range(vno):
        j = skip*i
        #print j
        temp.append(circle[j])
    temp.append(temp[0])
    #print len(temp)
    return temp

def event_to_animate_svg(oo, bgfile, filename_prefix=None, title="The event points", xmin=None, ymin=None, xmax=None, ymax=None, scale=30.):
    ###
    if filename_prefix is None:
        filename_prefix = 'temp_output/temp'

    dirname = os.path.dirname(filename_prefix)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    svgpanfile = dirname+"/SVGPan.js"
    if not os.path.exists(svgpanfile):
        svgpanfile_ori = os.path.dirname(__file__)+"/vmapper/SVGPan.js"
        copyfile(svgpanfile_ori, svgpanfile)

    filename = filename_prefix+'_event_map.svg'
    ###

    node_df = oo.node_df

    iis = node_df.node_id.tolist()
    xxs = node_df.xx.tolist()
    yys = node_df.yy.tolist()
    if xmin is None:
        xmin = min(xxs)
        ymin = min(yys)
        xmax = max(xxs)
        ymax = max(yys)
    boxwidth = xmax-xmin
    boxheight = ymax-ymin
    buffer_x = boxwidth*0.025
    buffer_y = boxheight*0.025
    tts = node_df.time.tolist()
    tt0 = min(tts)-1

    param_dict = {"title":title, "title_in_page":title, "xmin":xmin-buffer_x, "ymin":ymin-buffer_y, "boxwidth":boxwidth+2*buffer_x, "boxheight":boxheight+2*buffer_y, "bgcolor":"#000000", "textcolor":"#FFFFFF"}

    scene = svg_making.simple_scene(param_dict)
    bg,bbox = _get_background(bgf=bgfile, layername='background')
    if not(bg is None):
        scene.add_simpleLayer(bg)
        x1, y1, x2, y2 = bbox
        bw = x2-x1
        bh = y2-y1
        bx = bw*0.025
        by = bh*0.025
        scene.update_param({'xmin':x1-bx,'boxwidth':bw+2*bx,'ymin':y1-by,'boxheight':bh+2*by})

    for ii,xx,yy,tt in zip(iis,xxs,yys,tts):
        t0 = tt - tt0
        t1 = t0
        t2 = t1 + 1
        size = 1
        item = svg.Appear_Circle(center=(xx,yy), radius=size*scale, color=(255,0,0), idd=ii, t0=t0, t1=t1, t2=t2)
        scene.add_simpleLayer([item])
    outputText = scene.render()
    f = open(filename, 'w')
    f.write(outputText.encode("utf-8"))
    f.close()

def pg_to_animate_svg(oo, pgraph, bgfile, filename_prefix=None, title="The progression of subcluster", xmin=None, ymin=None, xmax=None, ymax=None, scale=30.):
    ###
    if filename_prefix is None:
        filename_prefix = 'temp_output/temp'

    dirname = os.path.dirname(filename_prefix)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    svgpanfile = dirname+"/SVGPan.js"
    if not os.path.exists(svgpanfile):
        svgpanfile_ori = os.path.dirname(__file__)+"/vmapper/SVGPan.js"
        copyfile(svgpanfile_ori, svgpanfile)

    filename = filename_prefix+'_prog_map.svg'
    ###

    node_df = oo.node_df
    xxs = node_df.xx.tolist()
    yys = node_df.yy.tolist()
    if xmin is None:
        xmin = min(xxs)
        ymin = min(yys)
        xmax = max(xxs)
        ymax = max(yys)
    boxwidth = xmax-xmin
    boxheight = ymax-ymin
    buffer_x = boxwidth*0.025
    buffer_y = boxheight*0.025
    tts = node_df.time.tolist()
    tt0 = min(tts)-1

    param_dict = {"title":title, "title_in_page":title, "xmin":xmin-buffer_x, "ymin":ymin-buffer_y, "boxwidth":boxwidth+2*buffer_x, "boxheight":boxheight+2*buffer_y, "bgcolor":"#000000", "textcolor":"#FFFFFF"}

    scene = svg_making.simple_scene(param_dict)
    bg,bbox = _get_background(bgf=bgfile, layername='background')
    if not(bg is None):
        scene.add_simpleLayer(bg)
        x1, y1, x2, y2 = bbox
        bw = x2-x1
        bh = y2-y1
        bx = bw*0.025
        by = bh*0.025
        scene.update_param({'xmin':x1-bx,'boxwidth':bw+2*bx,'ymin':y1-by,'boxheight':bh+2*by})

    for chain in pgraph.isolated_clusters:
        for cluster in chain:
            idd = cluster.cluster_id
            xx = cluster.xcor
            yy = cluster.ycor
            t0 = cluster.time_start - tt0
            t1 = cluster.time_median - tt0
            t2 = cluster.time_stop - tt0
            size = cluster.cluster_size

            item = svg.Appear_Circle(center=(xx,yy), radius=size*scale, color=(255,0,0), idd=idd, t0=t0, t1=t1, t2=t2)

            scene.add_simpleLayer([item])

    for chain in pgraph.long_chains:
        for cluster in chain:
            idd = cluster.cluster_id
            xx = cluster.xcor
            yy = cluster.ycor
            t0 = cluster.time_start - tt0
            t1 = cluster.time_median - tt0
            t2 = cluster.time_stop - tt0
            size = cluster.cluster_size

            item = svg.Appear_Circle(center=(xx,yy), radius=size*scale, color=(255,0,0), idd=idd, t0=t0, t1=t1, t2=t2)

            scene.add_simpleLayer([item])

    for link in pgraph.pg.progress:
        idda = link.origin.cluster_id
        xxa = link.origin.xcor
        yya = link.origin.ycor
        t0a = link.origin.time_start - tt0
        t1a = link.origin.time_median - tt0
        t2a = link.origin.time_stop - tt0

        iddb = link.destination.cluster_id
        xxb = link.destination.xcor
        yyb = link.destination.ycor
        t0b = link.destination.time_start - tt0
        t1b = link.destination.time_median - tt0
        t2b = link.destination.time_stop - tt0

        idd=str(idda)+'-'+str(iddb)
        start = (xxa,yya)
        end = (xxb,yyb)
        t0=t0a
        t1=t0b
        item = svg.animated_line(idd=idd, start=start, end=end, t0=t0, t1=t1)
        scene.add_simpleLayer([item])
    outputText = scene.render()
    f = open(filename, 'w')
    f.write(outputText.encode("utf-8"))
    f.close()

def pg_to_static_svg(oo, pgraph, bgfile, filename_prefix=None, title="The static view of subcluster", xmin=None, ymin=None, xmax=None, ymax=None, shape='ellipse', widthscale=1.):
    ###
    if filename_prefix is None:
        filename_prefix = 'temp_output/temp'

    dirname = os.path.dirname(filename_prefix)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    svgpanfile = dirname+"/SVGPan.js"
    if not os.path.exists(svgpanfile):
        svgpanfile_ori = os.path.dirname(__file__)+"/vmapper/SVGPan.js"
        copyfile(svgpanfile_ori, svgpanfile)

    filename = filename_prefix+'_prog_map_static.svg'
    ###===
    node_df = oo.node_df

    iis = node_df.node_id.tolist()
    xxs = node_df.xx.tolist()
    yys = node_df.yy.tolist()
    if xmin is None:
        xmin = min(xxs)
        ymin = min(yys)
        xmax = max(xxs)
        ymax = max(yys)
    boxwidth = xmax-xmin
    boxheight = ymax-ymin
    buffer_x = boxwidth*0.025
    buffer_y = boxheight*0.025
    tts = node_df.time.tolist()
    tt0 = min(tts)-1

    param_dict = {"title":title, "title_in_page":title, "xmin":xmin-buffer_x, "ymin":ymin-buffer_y, "boxwidth":boxwidth+2*buffer_x, "boxheight":boxheight+2*buffer_y, "bgcolor":"#000000", "textcolor":"#FFFFFF"}

    noise = []
    for anode in pgraph.sg.nodes:
        clid = anode.cluster_id
        #tt = anode.time
        if clid == -1:
            noise.append(anode)

    solo = []
    chain = []
    for ac in pgraph.pg.clusters:
        if ac.chain_id == -1:
            solo.append(ac)
        else:
            chain.append(ac)

    scene = svg_making.simple_scene(param_dict)
    bg,bbox = _get_background(bgfile, layername='background')
    if not(bg is None):
        scene.add_simpleLayer(bg)
        x1, y1, x2, y2 = bbox
        bw = x2-x1
        bh = y2-y1
        bx = bw*0.025
        by = bh*0.025
        scene.update_param({'xmin':x1-bx,'boxwidth':bw+2*bx,'ymin':y1-by,'boxheight':bh+2*by})

    for anode in noise:
        xx = anode.xcor
        yy = anode.ycor
        ii = str(anode.idd)
        tt = anode.time
        size = 1
        circle = []
        if shape=='ellipse':
            circle = _get_ellipse([anode], vno=16)
        else:
            circle = _get_circle([anode], vno=16)
        shpstr = svg.get_pathstring(vertexlist=circle)
        pid = 'noise_'+ii
        item = svg.linePath(shpstr, path_id=pid, stroke_width=8*widthscale, stroke_color=(50,50,50), fill_color=(200,200,200), stroke_opacity=1.0)
        scene.add_simpleLayer([item])

    for acluster in solo:
        xx = acluster.xcor
        yy = acluster.ycor
        tt = acluster.time_median
        size = acluster.cluster_size
        node_list = acluster.node_list
        circle = []
        if shape=='ellipse':
            circle = _get_ellipse(node_list, vno=16)
        else:
            circle = _get_circle(node_list, vno=16)
        shpstr = svg.get_pathstring(vertexlist=circle)
        pid = 'noise_'+ii
        item = svg.linePath(shpstr, path_id=pid, stroke_width=8*widthscale, stroke_color=(200,200,200), fill_color=(250,250,0), stroke_opacity=0.7)
        scene.add_simpleLayer([item])

    for link in pgraph.pg.progress:
        idda = link.origin.cluster_id
        xxa = link.origin.xcor
        yya = link.origin.ycor
        t0a = link.origin.time_start - tt0
        t1a = link.origin.time_median - tt0
        t2a = link.origin.time_stop - tt0

        iddb = link.destination.cluster_id
        xxb = link.destination.xcor
        yyb = link.destination.ycor
        t0b = link.destination.time_start - tt0
        t1b = link.destination.time_median - tt0
        t2b = link.destination.time_stop - tt0

        idd=str(idda)+'-'+str(iddb)
        start = (xxa,yya)
        end = (xxb,yyb)
        t0=t0a
        t1=t0b
        item = svg.Line(start, end, strokewidth=60*widthscale)
        scene.add_simpleLayer([item])

    for acluster in chain:
        xx = acluster.xcor
        yy = acluster.ycor
        tt = acluster.time_median
        size = acluster.cluster_size
        node_list = acluster.node_list
        circle = []
        if shape=='ellipse':
            circle = _get_ellipse(node_list, vno=16)
        else:
            circle = _get_circle(node_list, vno=16)
        shpstr = svg.get_pathstring(vertexlist=circle)
        pid = 'noise_'+ii
        item = svg.linePath(shpstr, path_id=pid, stroke_width=8*widthscale, stroke_color=(200,200,200), fill_color=(255,0,0), stroke_opacity=0.7)
        scene.add_simpleLayer([item])
    ### add legend text
    scene.add_text(svg_making.add_textbox())

    ### rendering
    outputText = scene.render()

    ### outputing
    f = open(filename, 'w')
    f.write(outputText.encode("utf-8"))
    f.close()
    ###===

def pg_to_morph_svg(oo, pgraph, bgfile, filename_prefix=None, title="The progression of subcluster", xmin=None, ymin=None, xmax=None, ymax=None, shape='ellipse', scale=30.):
    ###
    if filename_prefix is None:
        filename_prefix = 'temp_output/temp'

    dirname = os.path.dirname(filename_prefix)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    svgpanfile = dirname+"/SVGPan.js"
    if not os.path.exists(svgpanfile):
        svgpanfile_ori = os.path.dirname(__file__)+"/vmapper/SVGPan.js"
        copyfile(svgpanfile_ori, svgpanfile)

    filename = filename_prefix+'_morph_map.svg'
    ####


    node_df = oo.node_df

    iis = node_df.node_id.tolist()
    xxs = node_df.xx.tolist()
    yys = node_df.yy.tolist()
    if xmin is None:
        xmin = min(xxs)
        ymin = min(yys)
        xmax = max(xxs)
        ymax = max(yys)
    boxwidth = xmax-xmin
    boxheight = ymax-ymin
    buffer_x = boxwidth*0.025
    buffer_y = boxheight*0.025
    tts = node_df.time.tolist()
    tt0 = min(tts)-1

    param_dict = {"title":title, "title_in_page":title, "xmin":xmin-buffer_x, "ymin":ymin-buffer_y, "boxwidth":boxwidth+2*buffer_x, "boxheight":boxheight+2*buffer_y, "bgcolor":"#000000", "textcolor":"#FFFFFF"}

    time_noise = {}
    time_solo = {}
    time_chain = {}
    for anode in pgraph.sg.nodes:
        clid = anode.cluster_id
        tt = anode.time
        chid = anode.chain_id
        if clid == -1:
            if tt not in time_noise:
                time_noise[tt] = []
            time_noise[tt].append(anode)
        elif chid == -1:
            if clid not in time_solo:
                time_solo[clid] = {}
            if tt not in time_solo[clid]:
                time_solo[clid][tt] = []
            time_solo[clid][tt].append(anode)
        else:
            if chid not in time_chain:
                time_chain[chid] = {}
            if clid not in time_chain[chid]:
                time_chain[chid][clid] = {}
            if tt not in time_chain[chid][clid]:
                time_chain[chid][clid][tt] = []
            time_chain[chid][clid][tt].append(anode)

    ctime_noise = {}
    ctime_solo = {}
    ctime_chain = {}
    for tt, node_list in time_noise.iteritems():
        for tx in range(pgraph.T1):
            tt_delay = tt+tx
            if tt_delay not in ctime_noise:
                ctime_noise[tt_delay] = []
            ctime_noise[tt_delay].extend(node_list)

    for clid, cluster_dict in time_solo.iteritems():
        for tt, node_list in cluster_dict.iteritems():
            for tx in range(pgraph.T1):
                tt_delay = tt+tx
                if clid not in ctime_solo:
                    ctime_solo[clid] = {}
                if tt_delay not in ctime_solo[clid]:
                    ctime_solo[clid][tt_delay] = []
                ctime_solo[clid][tt_delay].extend(node_list)

    for chid, chain_dict in time_chain.iteritems():
        for clid, cluster_dict in chain_dict.iteritems():
            for tt, node_list in cluster_dict.iteritems():
                for tx in range(pgraph.T1):
                    tt_delay = tt+tx
                    if chid not in ctime_chain:
                        ctime_chain[chid] = {}
                    if clid not in ctime_chain[chid]:
                        ctime_chain[chid][clid] = {}
                    if tt_delay not in ctime_chain[chid][clid]:
                        ctime_chain[chid][clid][tt_delay] = []
                    ctime_chain[chid][clid][tt_delay].extend(node_list)

    clusters_dict = {}
    for ac in pgraph.pg.clusters:
        clid = ac.cluster_id
        clusters_dict[clid] = ac

    scene = svg_making.simple_scene(param_dict)
    bg,bbox = _get_background(bgfile, layername='background')
    if not(bg is None):
        scene.add_simpleLayer(bg)
        x1, y1, x2, y2 = bbox
        bw = x2-x1
        bh = y2-y1
        bx = bw*0.025
        by = bh*0.025
        scene.update_param({'xmin':x1-bx,'boxwidth':bw+2*bx,'ymin':y1-by,'boxheight':bh+2*by})

    for tt,node_list in ctime_noise.iteritems():
        for anode in node_list:
            xx = anode.xcor
            yy = anode.ycor
            ii = 'n_'+str(anode.idd)
            t0 = tt - tt0
            t1 = t0
            t2 = t1 + 1
            size = 1
            item = svg.Appear_Circle(center=(xx,yy), radius=size*scale, color=(200,200,200), idd=ii, t0=t0, t1=t1, t2=t2)
            scene.add_simpleLayer([item])


    temp_solo = {}
    for clid, ttcluster_dict in ctime_solo.iteritems():
        #this_cluster = clusters_dict[clid]
        if clid not in temp_solo:
            temp_solo[clid] = {}
        for tt, node_list in ttcluster_dict.iteritems():
            circle = []
            if shape=='ellipse':
                circle = _get_ellipse(node_list, vno=16)
            else:
                circle = _get_circle(node_list, vno=16)
            tt2 = tt - tt0
            temp_solo[clid][tt2] = circle
        temp_t = sorted(temp_solo[clid].keys())
        tot = max(temp_t)-min(temp_t)
        mint = min(temp_t)
        initv = temp_solo[clid][mint]
        initss = svg.get_pathstring(vertexlist=initv)
        ii = 'cl_'+str(clid)
        item = svg.Morphing_Shapes(idd=ii, init_str=initss, begin=mint, total_time=tot, fading_duration=1, color=(250,250,0))
        for i in temp_t:
            #if i!=mint:
            t0 = i-mint
            tx = t0/float(tot)
            this_c = temp_solo[clid][i]
            ss = svg.get_pathstring(vertexlist=this_c)
            item.add_shape(vertexstr=ss, tt=tx)
        scene.add_simpleLayer([item])

    temp_chain = {}
    for chid, ttchain_dict in ctime_chain.iteritems():
        for clid, ttcluster_dict in ttchain_dict.iteritems():
            if chid not in temp_chain:
                temp_chain[chid] = {}
            if clid not in temp_chain[chid]:
                temp_chain[chid][clid] = {}
            for tt, node_list in ttcluster_dict.iteritems():
                circle = []
                if shape=='ellipse':
                    circle = _get_ellipse(node_list, vno=16)
                else:
                    circle = _get_circle(node_list, vno=16)
                tt2 = tt - tt0
                temp_chain[chid][clid][tt2] = circle
                #break
            #break
            temp_t = sorted(temp_chain[chid][clid].keys())
            tot = max(temp_t)-min(temp_t)
            mint = min(temp_t)
            initv = temp_chain[chid][clid][mint]
            initss = svg.get_pathstring(vertexlist=initv)
            ii = 'cl_'+str(clid)
            item = svg.Morphing_Shapes(idd=ii, init_str=initss, begin=mint, total_time=tot, fading_duration=1, color=(250,0,0))
            for i in temp_t:
                #if i!=mint:
                t0 = i-mint
                tx = t0/float(tot)
                this_c = temp_chain[chid][clid][i]
                ss = svg.get_pathstring(vertexlist=this_c)
                item.add_shape(vertexstr=ss, tt=tx)
            scene.add_simpleLayer([item])

    ### add legend text
    scene.add_text(svg_making.add_textbox())

    ### rendering
    outputText = scene.render()

    ### outputing
    f = open(filename, 'w')
    f.write(outputText.encode("utf-8"))
    f.close()
"""



# -*- coding: utf-8 -*-

#import .utils.exporter as exporter
from .utils import oo, to_csv, to_shpfile, to_anime

class Outputs(object):
    def __init__(self, progression):
        #self.oo = exporter.Output_Objects(progression)
        self.oo = oo(progression)

    def to_csv(self, filename_prefix):
        #exporter.to_csv(self.oo, filename_prefix)
        to_csv(self.oo, filename_prefix)

    def to_shpfile(self, filename_prefix, crs, vno, dev_scale):
        #exporter.to_shpfile(self.oo, filename_prefix, crs, vno, dev_scale)
        to_shpfile(self.oo, filename_prefix, crs, vno, dev_scale)

    def to_anime(self, filename_prefix, bg_polys, bbox):
        #exporter.to_anime(self.oo, filename_prefix, bg_polys, bbox)
        to_anime(self.oo, filename_prefix, bg_polys, bbox)
    """
    def to_svgs(self, filename_prefix, pgraph, bgfile, bbox, scale, widthscale):
        exporter.to_svgs_all(self.oo, filename_prefix, pgraph, bgfile, bbox, scale, widthscale)

    def to_svg_event(self, bgfile, filename_prefix, title, bbox, scale):
        exporter.to_svg_event(self.oo, bgfile, filename_prefix, title=title, bbox=bbox, scale=scale)

    def to_svg_animate(self, pgraph, bgfile, filename_prefix, title, bbox, scale):
        exporter.to_svg_animate(self.oo, pgraph, bgfile, filename_prefix, title=title, bbox=bbox, scale=scale)

    def to_svg_static(self, pgraph, bgfile, filename_prefix, title, bbox, shape, widthscale):
        exporter.to_svg_static(self.oo, pgraph, bgfile, filename_prefix, title=title, bbox=bbox, shape=shape, widthscale=widthscale)

    def to_svg_morph(self, pgraph, bgfile, filename_prefix, title, bbox, shape, scale):
        exporter.to_svg_morph(self.oo, pgraph, bgfile, filename_prefix, title=title, bbox=bbox, shape=shape, scale=scale)
    """



# -*- coding: utf-8 -*-

import os

def to_csv(oo, filename_prefix=None):
    # oo == Output_Objects
    if filename_prefix is None:
        filename_prefix = 'temp_output/temp'

    dirname = os.path.dirname(filename_prefix)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    filename1 = filename_prefix+"_nodes.csv"
    filename2 = filename_prefix+"_clusters.csv"
    filename3 = filename_prefix+"_proglinks.csv"
    filename4 = filename_prefix+"_summary.csv"
    filename5 = filename_prefix+"_slinks_final.csv"
    filename6 = filename_prefix+"_cluster_pairs.csv"
    
    oo.node_df.to_csv(filename1, index_label='ind')
    oo.cluster_df.to_csv(filename2, index_label='ind')
    oo.progress_df.to_csv(filename3, index_label='ind')
    oo.summary_df.to_csv(filename4, index_label='ind')
    oo.final_slinks_df.to_csv(filename5, index_label='ind')
    oo.final_nlinks_df.to_csv(filename6, index_label='ind')



# -*- coding: utf-8 -*-

## node file

class Node():
    ## incident point in the data
    def __init__(self, idd, time, xcor, ycor):
        self.idd = idd
        self.time = time
        self.xcor = xcor
        self.ycor = ycor
        self.neighbors = []
        self.neighbor_nodes = []
        self.neighbors_final = []
        self.incoming = []
        #self.ori_weight = None # the combine risks of incoming
        #self.ori_rate = None # the combine risk / total combine risks
        self.outgoing = []
        self.cluster_id = None
        self.chain_id = None
        self.origin_pos_dict = {}

    def add_neighbor(self, a_nlink):
        self.neighbor_nodes.append(a_nlink.other_end(self))
        self.neighbors.append(a_nlink)

    def check_isneighbor(self, anode):
        if anode in self.neighbor_nodes:
            return True
        else:
            return False

    def add_incoming(self, a_slink):
        self.incoming.append(a_slink)

    def add_outgoing(self, a_slink):
        self.outgoing.append(a_slink)

    def add_finalnlink(self, a_nlink):
        self.neighbors_final.append(a_nlink)

    def set_clusterid(self, cid):
        self.cluster_id = cid

    def set_chain_id(self, cid):
        self.chain_id = cid

    def set_slink_orp(self, slink, oripos):
        self.origin_pos_dict[slink.origin.idd] = (slink, oripos)



# -*- coding: utf-8 -*-

## the shifting graph object,
## compose by nodes, neighboring pair, and shifting link

# import dependencies
import time
from random import randint
from scipy import spatial
from scipy.stats import norm
from numpy import std as nstd
# import local objects
from .node import Node
from .neighbor_pair import Neighbor_Pair
from .shifting_link import Shifting_Link

class Shift_Graph():
    ## the total picture of all nodes, neighbor_pair, and shifting_link
    def __init__(self):
        self.nodes = []
        self.isolated_nodes = []
        self.nlinks = []
        self.slinks = []
        self.nlinks_final = []
        self.slinks_final = []
        self.spatial_neighbor_list = None
        self.spatial_neighbor_dicts = None
        self.critical_value = None

    def add_node(self, idd, dd, xx, yy):
        anode = Node(idd, dd,xx,yy)
        self.nodes.append(anode)
        #self.nodes_dict[dd] = anode

    def _add_nlink(self, n1, n2):
        if not n1.check_isneighbor(n2):
            a_nlink = Neighbor_Pair(n1, n2)
            self.nlinks.append(a_nlink)
            n1.add_neighbor(a_nlink)
            n2.add_neighbor(a_nlink)

    def _add_slink(self, n1, n2, s_radius, T1, T2):
        a_slink = Shifting_Link(n1, n2, s_radius, T1, T2)
        self.slinks.append(a_slink)
        n2.add_incoming(a_slink)
        n1.add_outgoing(a_slink)

    def prepare(self, s_radius, T2):
        xlist = []
        ylist = []
        for n in self.nodes:
            xlist.append(n.xcor)
            ylist.append(n.ycor)
        pts = list(zip(xlist, ylist))
        #print('debug this:', len(pts))
        tree = spatial.KDTree(pts)
        nlist = tree.query_ball_point(x=pts, r=s_radius)
        self.spatial_neighbor_list = []
        self.spatial_neighbor_dicts = []
        for i in range(len(nlist)):
            nns = nlist[i]
            nns.remove(i)
            self.spatial_neighbor_list.append(nns)
            temp_dict = {}
            a_node = self.nodes[i]
            a_t = a_node.time
            for j in nns:
                b_node = self.nodes[j]
                b_t = b_node.time
                if( b_t <= (a_t + T2)) and (b_t >= a_t):
                    if b_t not in temp_dict:
                        temp_dict[b_t] = []
                    temp_dict[b_t].append(b_node)
            self.spatial_neighbor_dicts.append(temp_dict)


    def find_links(self, s_radius, T1, T2):
        ## ===============
        print('find links - start')
        #print 'calculate passing possibility - start'
        ## ===============
        nodes = self.nodes
        spatial_neighbor_list = self.spatial_neighbor_list

        for i in range(len(nodes)):
            a_node = nodes[i]
            t_i = a_node.time
            this_sns = self.spatial_neighbor_dicts[i]
            for tt in range(T1):
                tx = tt + t_i
                if tx in this_sns:
                    #print tx
                    #x = [ self._add_nlink(a_node, b) for b in this_sns[tx] ]
                    for b_node in this_sns[tx]:
                        self._add_nlink(a_node,b_node)
            for tt in range(T1,T2+1):
                tx = tt + t_i
                if tx in this_sns:
                    #x = [ self._add_slink(a_node, b, s_radius, T1, T2) for b in this_sns[tx] ]
                    for b_node in this_sns[tx]:
                        self._add_slink(a_node, b_node, s_radius, T1, T2)

        ## ===============
        print('find links - stop')
        print('calculate passing possibility - start')
        ## ===============
        self._calculate_passing_possibility()
        ## ===============
        print('calculate passing possibility - stop')
        print('calculate propensity - start')
        ## ===============
        self._calculate_propensity() ## maybe this part slow
        ## ===============
        print('calculate propensity - stop')
        #print 'calculate propensity - start'
        ## ===============

    def _calculate_passing_possibility(self):
        nodes = self.nodes
        origins = []
        for n in nodes:
            incomings = n.incoming
            if len(incomings)>0:
                #print len(incomings)
                iws = [i.combine_risk for i in incomings]
                tot = float(sum(iws))
                ori_rate = [w/tot for w in iws]
                #n.set_ori_weight(iws)
                #n.set_ori_rate(ori_rate)
                for ll,oo in zip(incomings, ori_rate):
                    ll.set_ori_pos(oo)
                    n.set_slink_orp(ll,oo)
            else:
                if len(n.outgoing)<=0:
                    self.isolated_nodes.append(n)
                else:
                    origins.append(n)
        #print "isolated nodes count:", str(len(self.isolated_nodes))
        #print "origins nodes count:", str(len(origins))

    def _calculate_propensity(self):
        nlinks = self.nlinks
        for a_link in nlinks:
            a_link.calculate_cop()
            #break

    def bootstraping(self, resample_time, confidence_level):
        sample_list = []
        #print nsize
        for i in range(resample_time):
            boot_list = []
            nsize = len(self.nlinks)
            if nsize>0:
                for j in range(nsize):
                    k = randint(0,nsize-1)
                    this_nlink = self.nlinks[k]
                    if not (this_nlink.cop is None):
                        boot_list.append(this_nlink.cop)
                    else:
                        boot_list.append(0)
                nboot = sum(boot_list) / float(len(boot_list))
                sample_list.append(nboot)
            else:
                nboot = None
        """
        cv = 0
        if confidence_level == 0.80:
            cv = 1.28
        elif confidence_level == 0.85:
            cv = 1.44
        elif confidence_level == 0.90:
            cv = 1.64
        elif confidence_level == 0.95:
            cv = 1.96
        elif confidence_level == 0.99:
            cv = 2.566
        """
        p = 1.-(1.-float(confidence_level))/2.
        cv = norm.ppf(p)
        if len(sample_list)>0:
            cop_mean = sum(sample_list)/float(len(sample_list))
            cop_std = nstd(sample_list)
            critical_value = cop_mean + cv*cop_std
        else:
            critical_value = None

        self.critical_value = critical_value
        return critical_value

    def skip_bootstraping(self, critical_value):
        self.critical_value = critical_value
        return critical_value

    def finalize_nlinks(self):
        ### filtering the nlinks that is lower than critical value, save the "cluster link" to final nlink set
        if self.critical_value is None:
            print("error: critical value is None")
            return
        #check = 0
        for nlin in self.nlinks:
            if not (nlin.cop is None):
                #check=check+1
                if nlin.cop >= self.critical_value:
                    self.nlinks_final.append(nlin)
        #print 'check: ', str(check)
        for anode in self.nodes:
            for nl in anode.neighbors:
                if not (nl.cop is None):
                    if nl.cop >= self.critical_value:
                        anode.add_finalnlink(nl)

    def finalize_slinks(self):
        # filtering slink, save to final slink set
        #print 'final nlinks count: ', str(len(self.nlinks_final))
        s_links_final = []
        origin_set = []
        for anode in self.nodes:
            # for those node with no cluster link neighbor
            # get the most probable shifting link that might happen
            if len(anode.neighbors_final)<=0:
                s_incoming = anode.incoming
                if len(s_incoming)>0:
                    s_poss = [x.ori_possibility for x in s_incoming]
                    M_inc = None
                    M_pos = 0
                    temp_dict = {}
                    for sl, sp in zip(s_incoming, s_poss):
                        if sp not in temp_dict:
                            temp_dict[sp] = []
                        temp_dict[sp].append(sl)
                        #if sp > M_pos:
                        #    M_inc = sl
                        #    M_pos = sp
                    M_pos = max(temp_dict.keys())
                    M_inc = temp_dict[M_pos]
                    s_links_final.extend(M_inc)
                else:
                    origin_set.append(anode)
            else:
                # for those with cluster link
                # get the highest common origin node
                # save the slinks from that node to the both end of the nlink
                n_links = anode.neighbors_final
                for nl in n_links:
                    if nl.max_cop is not None:
                        s_links_final.extend(nl.link1_max)
                        s_links_final.extend(nl.link2_max)
        #s_links_final2 = [x for x in s_links_final if x is not None]
        self.slinks_final = list(set(s_links_final))



# -*- coding: utf-8 -*-
#from __future__ import absolute_import

#import time

import pandas as pd
import geopandas as gpd

from .core_objs import Shift_Graph, Progression_Graph
from .utils import wqu
#import core_objs as core
#from .utils.weightedquickunion import GenericWeightedQuickUnion as wqu
#from utils.exporter import exporter as exporting

class Diffusion_Progress(object):
    def __init__(self, ptdataframe, pts_setting, s_radius, T1, T2, resample_time, confidence_level, critical_value, crs=None):
        self.T1 = T1
        self.T2 = T2
        #pts = gpd.read_file(pointfile)
        pts = ptdataframe
        self.sg = Shift_Graph()
        self.crs = crs
        #self.points = []
        ## ===============
        print('construction of shifting graph - start')
        ## ===============
        pts_setting_default = {'xcor':'xcor', 'ycor':'ycor', 'time':'time'}
        if pts_setting is None:
            pts_setting = {'xcor':'xcor', 'ycor':'ycor', 'time':'time'}
        else:
            for k,v in pts_setting_default.items():
                if not(k in pts_setting):
                    pts_setting[k] = v
        self.__get_points(pts, pts_setting)

        ## ===============
        print('construction of shifting graph - stop')
        print('making links - start')
        ## ===============
        self.__make_links(s_radius, T1, T2)

        ## ===============
        print('making links - stop')
        print('bootstraping - start')
        ## ===============

        if not(critical_value is None):
            self.critical_value = self.sg.skip_bootstraping(critical_value)
        else:
            self.critical_value = self.sg.bootstraping(resample_time, confidence_level)
        print('critical value is: ', self.critical_value)

        self.sg.finalize_nlinks() # must calculate before finalize slink
        self.sg.finalize_slinks() # must calculate after finalize nlink
        #print 'final nlinks count:', str(len(self.sg.nlinks_final))
        #print 'final slinks count:', str(len(self.sg.slinks_final))

        ## ===============
        print('bootstraping - stop')
        print('detection of subclusters - start')
        ## ===============

        self.pg = Progression_Graph()
        self.__subcluster_detection() # can run after nlinks are finalized
        self.__find_progression_links() # can run after slinks are finalized
        self.chain_list_full = None
        self.long_chains = None
        self.isolated_clusters = None
        self.__get_progression_chain()

        print('number of subcluster found:', str(len(self.pg.clusters)))

        ## component graph construction
        ## track progression
        #self.exporter = exporting.Export_Main_Result(self)

    def __get_points(self, pts, pts_setting):
        pts = pts.sort_values(by=pts_setting['time'])
        xs = []
        ys = []
        ts = pts[pts_setting['time']].tolist()
        if isinstance(pts, gpd.GeoDataFrame):
            geoms = pts.geometry.tolist()
            xs = [ p.x for p in geoms ]
            ys = [ p.y for p in geoms ]
            self.crs = pts.crs
        elif isinstance(pts, pd.DataFrame):
            xs = pts[pts_setting['xcor']].tolist()
            ys = pts[pts_setting['ycor']].tolist()
        i = 0
        for xcor,ycor,time in list(zip(xs,ys,ts)):
            #this_point = pts.iloc[i]
            #time = this_point[pts_setting['time']]
            #xcor = this_point[pts_setting['xcor']]
            #ycor = this_point[pts_setting['ycor']]
            self.sg.add_node(i, time, xcor, ycor)
            i+=1

    def __make_links(self, s_radius, T1, T2):
        self.sg.prepare(s_radius, T2)
        self.sg.find_links(s_radius, T1, T2)

    def __subcluster_detection(self):
        cluster_wqu = wqu(init_list=self.sg.nodes)
        for npair in self.sg.nlinks_final:
            cluster_wqu.union(npair.one, npair.two)
        clusters_list = cluster_wqu.get_currentcomponents(form='list')
        j = 0
        for i in range(len(clusters_list)):
            nodes = clusters_list[i]
            if len(nodes)>1:
                for anode in nodes:
                    anode.set_clusterid(j)
                self.pg.add_cluster_node(j, nodes)
                j = j + 1
            else:
                for anode in nodes:
                    anode.set_clusterid(-1)
        self.pg.finalize_clusters()

    def __find_progression_links(self):
        #clusters = self.pg.clusters
        for spair in self.sg.slinks_final:
            #if not (spair is None):
            s_ori = spair.origin.cluster_id
            s_des = spair.destination.cluster_id
            ori_possibility = spair.ori_possibility ###
            if (s_ori>=0) and (s_des>=0) and (s_ori!=s_des):
                #c1 = clusters[s_ori]
                #c2 = clusters[s_des]
                t1 = self.pg.clusters_dict[s_ori].time_start
                t2 = self.pg.clusters_dict[s_des].time_start
                if t1<t2:
                    self.pg.add_progressionlink(s_ori, s_des, ori_possibility)
        for ac in self.pg.clusters:
            ac.finalize_behaviors()

    def __get_progression_chain(self):
        chain_wqu = wqu(init_list=self.pg.clusters)
        for spair in self.pg.progress:
            chain_wqu.union(spair.origin, spair.destination)
        chain_list = chain_wqu.get_currentcomponents(form='list')
        self.chain_list_full = chain_list
        #print len(chain_list)
        sizes = []
        long_chains = []
        isol_chain = []
        j = 0
        for i in range(len(chain_list)):
            clusters = chain_list[i]
            if len(clusters) > 1:
                sizes.append(len(clusters))
                for ac in clusters:
                    ac.set_chain_id(j)
                    for anode in ac.node_list:
                        anode.set_chain_id(j)
                long_chains.append(clusters)
                j = j + 1
            else:
                for ac in clusters:
                    ac.set_chain_id(-1)
                    for anode in ac.node_list:
                        anode.set_chain_id(-1)
                isol_chain.append(clusters)
        self.long_chains = long_chains
        self.isolated_clusters = isol_chain
        #print sizes
        #print len(sizes)



# -*- coding: utf-8 -*-

## shifting link

from numpy import exp, sqrt

class Shifting_Link():
    ## connection between nodes near to each other in space, but between a range of T1 to T2 starting from the origin node's time to destination node
    def __init__(self, node1, node2, s_radius, T1, T2):
        self.origin = node1
        self.destination = node2
        dd = self._get_distance(node1, node2)
        tt = node2.time - node1.time
        self.spatial_risk = self._get_spatial_risk(dd, s_radius)
        self.temporal_risk = self._get_temporal_risk(tt, T1, T2)
        self.combine_risk = self.spatial_risk * self.temporal_risk
        self.ori_possibility = None

    def _get_distance(self, node1, node2):
        dx = node2.xcor-node1.xcor
        dy = node2.ycor-node1.ycor
        dd = sqrt(dx**2+dy**2)
        return dd

    def _get_spatial_risk(self, spatial_distance, s_radius):
        Rs = float((1 - spatial_distance/float(s_radius))**2)
        return Rs

    def _get_temporal_risk(self, temporal_distance, T1, T2):
        #Rt = float(exp(-(temporal_distance-(T1+T2)/2.0)**2/2.0))
        Rt = float(exp(-(temporal_distance-float(T1+T2)/2.0)**2/float(T2-T1))) # modified 2017/2/10
        return Rt

    def set_ori_pos(self, arate):
        self.ori_possibility = arate



# -*- coding: utf-8 -*-
from .weightedquickunion import GenericWeightedQuickUnion as wqu
from .exporter.output_objects import Output_Objects as oo
from .exporter.to_shpfile import to_shpfile as to_shpfile
from .exporter.to_csv import to_csv as to_csv
from .exporter.to_anime import to_anime as to_anime




## author: benny chin @ 2015/12/02 ##
## a generic class for weighted quick union implementation ##
# lists of string, integer are tested...
# a lookup dictionary is used for identifying item from the "simple list", which record the root of the item at the index

class GenericWeightedQuickUnion():
    def __init__(self, init_list=[]):
        self._roots = [] ## parent (position as in the init_list)
        self.lookuptable = {} ## dict(item = position in init_list)
        self.sizes = [] ## child size(position as in the init_list)
        self.item_list = [] ## item (position as in the init_list)
        for i in range(len(init_list)):
            self.lookuptable[init_list[i]] = i
            self._roots.append(i)
            self.sizes.append(1)
            self.item_list.append(init_list[i])

    def additem(self, item):
        j = len(self._roots)
        self.lookuptable[item] = j
        self._roots.append(j)
        self.sizes.append(1)
        self.item_list.append(item)

    def union(self, item_p, item_q):
        p = self.lookuptable[item_p]
        q = self.lookuptable[item_q]
        self._union(p, q)

    def _union(self, p, q):
        i = self._findroot(p)
        j = self._findroot(q)
        if i != j:
            sz_i = self.sizes[i]
            sz_j = self.sizes[j]
            if ( sz_i < sz_j ): # a speed up move
                self._roots[i] = j
                self.sizes[j] = sz_i+sz_j
            else:
                self._roots[j] = i
                self.sizes[i] = sz_i+sz_j

    def findroot(self, item_k):
        k = self.lookuptable[item_k]
        return self._findroot(k)

    def _findroot(self, k):
        r = k
        while (r != self._roots[r]):
            r = self._roots[r]
        self._roots[k] = r # a speed up move
        return r

    def isconnected(self, item_p, item_q):
        return self._isconnected(self.lookuptable[item_p], self.lookuptable[item_q])

    def _isconnected(self, p, q):
        return self._findroot(p) == self._findroot(q)

    def get_currentcomponents(self, form='list'):
        root_dict = {}
        for i in range(len(self._roots)):
            kp = self._roots[i] # parent of item_k
            item_k = self.item_list[i]
            top_root = self._findroot(kp)
            if top_root not in root_dict:
                root_dict[top_root] = []
            root_dict[top_root].append(item_k)
        if form == 'dict':
            return root_dict
        elif form == 'list':
            # re-index the id of component
            #components = {}
            components_list = []
            keys = sorted(root_dict.keys())
            for i in range(len(keys)):
                #components[i] = root_dict[keys[i]]
                components_list.append(root_dict[keys[i]])
            return components_list
        else:
            print("not reconized 'form', returning list")
            # re-index the id of component
            #components = {}
            components_list = []
            keys = sorted(root_dict.keys())
            for i in range(len(keys)):
                #components[i] = root_dict[keys[i]]
                components_list.append(root_dict[keys[i]])
            return components_list



## for testing purpose: items of objects
class node():
    def __init__(self, key):
        self.id = key

## for analyzing running time ##
class analysis_of_timerunning():
    def __init__(self):
        import time
        times = []
        klist = []
        k = 2
        for i in range(10):
            k = k*2
            klist.append(k)
            print(k)
            starttime = time.time()
            self.running(k)
            stoptime = time.time()
            times.append(stoptime - starttime)
        import matplotlib.pyplot as plt
        #fig = plt.Figure()
        #ax = fig.add_subplot(1,1,1)
        plt.loglog(klist, times, basex=2)
        plt.show()

    def running(self, k):
        #k=10000
        dlist = [ node(i) for i in range(k) ]
        wqu4 = GenericWeightedQuickUnion(dlist)
        for i in range(int(k/2-1)):
            wqu4.union( dlist[i], dlist[i+1] )
        for i in range(int(k/2-1)):
            wqu4.union( dlist[int(i+k/2)], dlist[int(i+k/2+1)] )
        print(wqu4.isconnected( dlist[0], dlist[k-1] ))
        wqu4.union(dlist[int(k/2-1)], dlist[int(k/2)])
        print(wqu4.isconnected( dlist[0], dlist[k-1] ))



if __name__ == '__main__':
    """
    print ('===========test 1===========')
    alist = ['a','b','c','d','e','f','g','h']
    wqu = GenericWeightedQuickUnion(alist)
    wqu.union('a','b')
    print wqu.isconnected('a','b')
    print wqu.isconnected('a','c')
    wqu.union('g','h')
    wqu.union('b','c')
    print wqu.isconnected('e','a')
    wqu.union('e','h')
    wqu.union('h','b')
    print wqu.isconnected('e','a')


    print ('===========test 2===========')
    blist = [1,2,3,4,5,6,7,8]
    wqu2 = GenericWeightedQuickUnion(blist)
    wqu2.union(1,2)
    print wqu2.isconnected(1,2)
    print wqu2.isconnected(1,8)
    wqu2.union(2,4)
    wqu2.union(8,7)
    print wqu2.isconnected(1,8)
    wqu2.union(4,7)
    print wqu2.isconnected(1,8)


    print ('===========test 3===========')
    clist = [ node(i) for i in range(10) ]
    wqu3 = GenericWeightedQuickUnion(clist)
    wqu3.union( clist[1], clist[2] )
    print wqu3.isconnected(clist[1], clist[2])
    print wqu3.isconnected(clist[1], clist[9])
    wqu3.union( clist[2], clist[5])
    wqu3.union( clist[5], clist[9])
    print wqu3.isconnected(clist[1], clist[9])
    xxx = node(11)
    wqu3.additem(xxx)
    print wqu3.isconnected(clist[1], xxx )
    wqu3.union( clist[1], xxx )
    print wqu3.isconnected(clist[1], xxx )
    """

    print ('===========test 4===========')
    #import sys
    #print(sys.version)
    analysis_of_timerunning()



# -*- coding: utf-8 -*-

## neigboring pair

class Neighbor_Pair():
    ## the connection between neighboring nodes (both time and space)
    def __init__(self, node1, node2):
        self.one = node1
        self.two = node2
        self.cop = None
        self.max_cop = None
        self.link1_max = None
        self.link2_max = None
        self.copmax_links = None

    def other_end(self, anode):
        if anode == self.one:
            return self.two
        else:
            return self.one

    def calculate_cop(self):
        # cop: common origin propensity
        Pcops = {}
        Pcops_list = []
        in_1 = self.one.origin_pos_dict
        in_2 = self.two.origin_pos_dict
        len1 = len(in_1)
        len2 = len(in_2)
        base_in = in_1
        check_in = in_2
        if len2<len1:
            base_in = in_2
            check_in = in_1
        for k in base_in.keys():
            if k in check_in:
                l1,orp1 = base_in[k]
                l2,orp2 = check_in[k]
                pcop = orp1*orp2
                Pcops_list.append(pcop)
                #Pcops[ori1] = pcop
                if pcop not in Pcops:
                    Pcops[pcop] = []
                Pcops[pcop].append((l1,l2))

        if len(Pcops)>0:
            cops = Pcops.keys()
            self.cop = sum(Pcops_list)
            #print cop_max == max(cops)
            self.max_cop = max(cops)
            self.copmax_links = Pcops[self.max_cop]
            self.link1_max = []
            self.link2_max = []
            for a,b in self.copmax_links:
                self.link1_max.append(a)
                self.link2_max.append(b)



# -*- coding: utf-8 -*-

import os
#from datetime import date,timedelta ## remove the date from output so nvm

import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Polygon, LineString, Point
from shapely.geometry.polygon import LinearRing # get ellipse

def to_shpfile(oo, filename_prefix=None, crs=None, vno=16, dev_scale=1.5):
    # oo == Output_Objects
    if filename_prefix is None:
        filename_prefix = 'temp_output/temp'

    dirname = os.path.dirname(filename_prefix)
    if not os.path.isdir(dirname):
        os.makedirs(dirname)

    to_nodes(oo, filename_prefix, crs)
    to_final_slinks(oo, filename_prefix, crs)
    to_final_nlinks(oo, filename_prefix, crs)
    to_eclipse(oo, filename_prefix, crs, vno, dev_scale)
    to_proglinks(oo, filename_prefix, crs)

def to_nodes(oo, filename_prefix, crs):
    outputfilename = filename_prefix+"_nodes.shp"
    ndf = oo.node_df
    xx = ndf.xx.tolist()
    yy = ndf.yy.tolist()
    geom = []
    for a,b in zip(xx,yy):
        p = Point(a,b)
        geom.append(p)
    node_gdf = gpd.GeoDataFrame(ndf, crs=crs, geometry=geom)
    node_gdf.to_file(filename=outputfilename, driver='ESRI Shapefile')

def to_eclipse(oo, filename_prefix, crs, vno, dev_scale):
    # start working
    outputfilename = filename_prefix+"_ellipse.shp"
    ndf = oo.node_df
    cdf = oo.cluster_df
    cdf.sort_values(by='chain_id', inplace=True)

    cids2 = cdf.chain_id.tolist()
    clids = cdf.cluster_id.tolist()
    #print cdf.head()
    #cltime = cdf.time_start.tolist()
    clx = cdf.xx.tolist()
    cly = cdf.yy.tolist()
    #basetime = date(y,4,1)
    cluster_ellipse = []
    #cluster_date = []
    for cl,mx,my in zip(clids,clx,cly):
        temp_df = ndf[ndf.cluster_id==cl]
        temp_x = temp_df.xx.tolist()
        temp_y = temp_df.yy.tolist()
        ellipse = get_ellipse(temp_x, temp_y, vno, dev_scale)
        poly = Polygon(ellipse)
        #the_date = basetime + timedelta(int(tt))
        #the_date2 = the_date.strftime('%Y/%m/%d')
        cluster_ellipse.append(poly)
        #cluster_date.append(the_date2)
    beh = cdf.behaviors.tolist()
    cdf['behaviors'] = [ str(b) for b in beh ]
    new_df = cdf.rename(columns={'cluster_size':'cls_size', 'time_median':'time_mdian'})
    #new_df['date'] = cluster_date

    geo_df = gpd.GeoDataFrame(new_df, crs=crs, geometry=cluster_ellipse)
    #print geo_df.head()
    geo_df.to_file(filename=outputfilename, driver='ESRI Shapefile')

def get_ellipse(xlist, ylist, vno, dev_scale):
    # vno is the number of vertices of the ellipse
    # dev_scale is the standard deviation scale
    size = len(xlist)
    xx = sum(xlist)/float(size)
    yy = sum(ylist)/float(size)

    xwave = [(a-xx) for a in xlist]
    ywave = [(b-yy) for b in ylist]
    xlist2 = [a**2 for a in xwave]
    ylist2 = [b**2 for a in ywave]
    xylist2 = [(a)*(b) for a,b in zip(xwave,ywave)]
    sumx2 = sum(xlist2)
    sumy2 = sum(ylist2)
    sumxy2 = sum(xylist2)
    A = sumx2 - sumy2
    B = np.sqrt((sumx2 - sumy2)**2 + 4*(sumxy2)**2)
    C = 2*sumxy2
    #circle = []
    if C!=0:
        theta = np.arctan((A+B)/C)

        cxlist = [a*np.cos(theta) for a in xwave]
        sylist = [b*np.sin(theta) for b in ywave]
        sxlist = [a*np.sin(theta) for a in xwave]
        cylist = [b*np.cos(theta) for b in ywave]
        partx = [(a-b)**2 for a,b in zip(cxlist,sylist)]
        party = [(a-b)**2 for a,b in zip(sxlist,cylist)]
        devx = np.sqrt(sum(partx)/float(size))*dev_scale
        devy = np.sqrt(sum(party)/float(size))*dev_scale
        #print devx

        n = 64
        t = np.linspace(0, 2*np.pi, n, endpoint=False)
        st = np.sin(t)
        ct = np.cos(t)
        sa = np.sin(theta)
        ca = np.cos(theta)
        p = np.empty((n, 2))
        p[:, 0] = xx + devx * ca * ct - devy * sa * st
        p[:, 1] = yy + devx * sa * ct + devy * ca * st
        circle = list(LinearRing(p).coords)
        #circle = LinearRing(p)
    else:
        dev = (np.sqrt(sum(xlist2)/float(size)) + 1)*dev_scale
        #print dev
        theta = 1
        n = 64
        t = np.linspace(0, 2*np.pi, n, endpoint=False)
        st = np.sin(t)
        ct = np.cos(t)
        sa = np.sin(theta)
        ca = np.cos(theta)
        p = np.empty((n, 2))
        p[:, 0] = xx + dev * ca * ct - dev * sa * st
        p[:, 1] = yy + dev * sa * ct + dev * ca * st
        circle = list(LinearRing(p).coords)
        #circle = LinearRing(p)

    #print circle
    skip = int(np.floor(len(circle[:-1])/float(vno-1)))
    temp = []
    #print vno
    for i in range(vno):
        j = skip*i
        #print j
        temp.append(circle[j])
    #temp.append(temp[0])
    #print len(temp)
    return circle



def to_proglinks(oo, filename_prefix, crs):
    outputfilename = filename_prefix+'_proglinks.shp'
    edf = oo.progress_df
    if not (edf.iloc[0]['x0']=='-'):
        geom = []
        for i in range(len(edf)):
            row = edf.iloc[i]
            x0 = row['x0']
            x1 = row['x1']
            y0 = row['y0']
            y1 = row['y1']
            line = LineString(((x0,y0),(x1,y1)))
            geom.append(line)

        #crs = '+proj=tmerc +lat_0=0 +lon_0=121 +k=0.9999 +x_0=250000 +y_0=0 +ellps=GRS80 +units=m +no_defs'

        geo_df = gpd.GeoDataFrame(edf, crs=crs, geometry=geom)
        #print geo_df.head()
        geo_df.to_file(filename=outputfilename, driver='ESRI Shapefile')
    else:
        pass


def to_final_slinks(oo, filename_prefix, crs):
    outputfilename = filename_prefix+'_slinks_final.shp'
    edf = oo.final_slinks_df
    if not (edf.iloc[0]['ooid']=='-'):
        geom = []
        for i in range(len(edf)):
            row = edf.iloc[i]
            x0 = row['oxcor']
            x1 = row['dxcor']
            y0 = row['oycor']
            y1 = row['dycor']
            line = LineString(((x0,y0),(x1,y1)))
            geom.append(line)
        geo_df = gpd.GeoDataFrame(edf, crs=crs, geometry=geom)
        #print geo_df.head()
        geo_df.to_file(filename=outputfilename, driver='ESRI Shapefile')
    else:
        pass


def to_final_nlinks(oo, filename_prefix, crs):
    outputfilename = filename_prefix+'_cluster_pairs.shp'
    edf = oo.final_nlinks_df
    if not (edf.iloc[0]['n1_id']=='-'):
        geom = []
        for i in range(len(edf)):
            row = edf.iloc[i]
            x0 = row['n1x']
            x1 = row['n2x']
            y0 = row['n1y']
            y1 = row['n2y']
            line = LineString(((x0,y0),(x1,y1)))
            geom.append(line)
        geo_df = gpd.GeoDataFrame(edf, crs=crs, geometry=geom)
        #print geo_df.head()
        geo_df.to_file(filename=outputfilename, driver='ESRI Shapefile')
    else:
        pass



 # -*- coding: utf-8 -*-

## main interface

#import calc_frame
from .Diffusion_Progress import Diffusion_Progress
from .output_func import Outputs

class Point_Diffusion(object):
    def __init__(self, ptdataframe, pts_setting=None, s_radius=500, T1=12, T2=27, resample_time=99, confidence_level=0.8, critical_value=None, crs=None):

        self.ptdataframe = ptdataframe
        self.pts_setting = pts_setting
        self.crs = crs
        self.s_radius = s_radius
        self.T1 = T1
        self.T2 = T2
        self.resample_time = resample_time
        self.confidence_level = confidence_level
        self.critical_value = critical_value

        self.progression = Diffusion_Progress( self.ptdataframe, self.pts_setting, self.s_radius, self.T1, self.T2, self.resample_time, self.confidence_level, self.critical_value, crs=crs )

        self.outputs = Outputs(self.progression)
        self.results = self.outputs.oo

    def to_csv(self, filename_prefix=None):
        self.outputs.to_csv(filename_prefix)

    def get_gdf(self, crs=None, vno=16, dev_scale=1.5):
        pass

    def to_shpfile(self, filename_prefix=None, crs=None, vno=16, dev_scale=1.5):
        if crs is None:
            crs = self.crs
        self.outputs.to_shpfile(filename_prefix, crs, vno, dev_scale)

    def to_anime(self, filename_prefix=None, bg_polys=[], bbox=None):
        self.outputs.to_anime(filename_prefix, bg_polys, bbox)
    """
    def to_svgs(self, filename_prefix=None, bgfile=None, bbox=None, scale=30., widthscale=1.):
        self.outputs.to_svgs(filename_prefix, self.progression, bgfile=bgfile, bbox=bbox, scale=scale, widthscale=widthscale)

    def to_svg_event(self, filename_prefix=None, bgfile=None, title="The event of points", bbox=None, scale=30.):
        self.outputs.to_svg_event(bgfile, filename_prefix, title=title, bbox=bbox, scale=scale)

    def to_svg_animate(self, filename_prefix=None, bgfile=None, title="The progression of subcluster", bbox=None, scale=30):
        self.outputs.to_svg_animate(self.progression, bgfile, filename_prefix, title=title, bbox=bbox, scale=scale)

    def to_svg_static(self, filename_prefix=None, bgfile=None, title="The static view of subcluster", bbox=None, shape='ellipse', widthscale=1.):
        self.outputs.to_svg_static(self.progression, bgfile, filename_prefix, title=title, bbox=bbox, shape=shape, widthscale=widthscale)

    def to_svg_morph(self, filename_prefix=None, bgfile=None, title="The progression of subcluster", bbox=None, shape='ellipse', scale=30.):
        self.outputs.to_svg_morph(self.progression, bgfile, filename_prefix, title=title, bbox=bbox, shape=shape, scale=scale)
    """



from setuptools import setup
from setuptools import find_packages

setup(
    name="tapitas",

    version="0.0.4",

    author="Benny Chin",
    author_email="wcchin.88@gmail.com",

    packages=['tapitas', 'tapitas.core_objs', 'tapitas.utils', 'tapitas.utils.exporter'],

    include_package_data=True,

    url="https://bitbucket.org/wcchin/TaPiTaS",

    license="LICENSE.txt",
    description="A data exploration and visualization algorithm for understanding diffusion process.",

    long_description=open("README.md").read(),

    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: GIS',

         'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 3.5',
    ],

    keywords='point diffusion subcluster progress',

    install_requires=[
        "numpy",
        "scipy",
        "pandas",
        "geopandas",
        "shapely",
        "descartes",
        "matplotlib",
        "seaborn",
    ],
)
